# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

class Measurement:
    """!
    Measured data mapping in DataIdentifier: value format with corresponding timestamp.
    """

    ## @var fields
    # Measurement fields.

    ## @var time
    # Measurement timestamp.

    def __init__(self, fields, time):
        """!
        Initiate measurement object.

        @param fields Maping {dataIdentifier: vaue}.
        @param time  Measurement timestamp.
        @throws AttributeError If length of fields is greater than 8.
        """
        if len(fields) > 8:
            raise AttributeError("Fields must be up to length 8, {} given".format(len(fields)))
        self.fields = fields
        self.time = time

    @classmethod
    def currentMeasurement(cls, fields):
        """!
        Build measurement object with current time.

        @param cls
        @param fields Maping {dataIdentifier: vaue}.
        """
        return cls(fields, datetime.datetime.now())

    def __str__(self):
        """!
        Convert object to string.

        @return String.
        """
        return "[{}] {}".format(self.time, self.fields)

    def __repr__(self):
        """!
        Convert object to representation string.

        @return representation string.
        """
        return "<{}>".format(self.__str__())

    def __len__(self):
        """!
        Get number of measurement fields.

        @return Number of fields.
        """
        return len(self.fields)

class MeasurementParamConverter:
    """!
    Convert data measurement into ThingSpeak fields for single channel.
    """

    ## @var dataFieldsMapping
    # Mapping of data fields.

    def __init__(self, dataFieldsMapping):
        """!
        Initiate MeasurementParamConverter object.

        @param dataFieldsMapping Object for mapping DataIdentifier object to
            ThingSpeak channel field {DataIdentifier: "field"}.
        """
        self.dataFieldsMapping = dataFieldsMapping

    def convert(self, measurement):
        """!
        Convert measurement.

        @param measurement Measurement object.
        @throws ConvertException If reguired measurement item is missing.
        """
        params = dict()
        for topicName in measurement.fields:
            fieldName = self.dataFieldsMapping[topicName]
            params[fieldName] = measurement.fields[topicName]
        return params

    def __str__(self):
        """!
        Convert object to string.

        @return String.
        """
        return "<Param Converter: {}>".format(str(self.dataFieldsMapping))

    def __repr__(self):
        """!
        Convert object to representation string.

        @return representation string.
        """
        return self.__str__()

class ConvertException(Exception):
    """!
    Conversion error.
    """



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum

class ChannelType(enum.Enum):
    """!
    Enumeration of channel types.
    """

    thingspeak = 0
    phant = 1

class Channel:
    """!
    ThingSpeak channel identification object.
    """

    ## @var channelType
    # Type of channel.

    ## @var name
    # Channel name.

    ## @var channelID
    # Channel ID or None.

    ## @var apiKey
    # Channel API write key.

    def __init__(self, channelType, name, channelID, apiKey):
        """!
        Initiate channel object.

        @param channelType ChannelType enumeration object.
        @param name Channel name.
        @param channelID Channel identification.
        @param apiKey Channel write API key.
        """
        self.channelType = channelType
        self.name = name
        self.channelID = channelID
        self.apiKey = apiKey

    def __hash__(self):
        """!
        Calculate hash from channel name and API key.

        @return Hash.
        """
        return hash((self.name, self.apiKey))

    def __str__(self):
        """!
        Convert Channel to string.

        @return String.
        """
        return "{} [Id: {}, Key: {}]".format(self.name, self.channelID, self.apiKey)

    def __repr__(self):
        """!
        Convert Channel to representation string.

        @return Representation string.
        """
        return "<{}>".format(self.__str__())

class ThingSpeakChannel(Channel):
    """!
    ThingSpeak channel identification object.
    """

    def __init__(self, name, channelID, apiKey):
        """!
        Initiate ThingSpeak channel object.

        @param name Channel name.
        @param channelID Channel identification.
        @param apiKey Channel write key.
        """
        Channel.__init__(self, ChannelType.thingspeak, name, channelID, apiKey)

class PhantChannel(Channel):
    """
    Phant channel identification object.
    """

    def __init__(self, name, channelID, apiKey):
        """!
        Initiate Phant channel object.

        @param name Channel name.
        @param channelID Channel identification.
        @param apiKey Channel write key.
        """
        self.checkChannelIdentification(channelID)
        Channel.__init__(self, ChannelType.phant, name, channelID, apiKey)

    def checkChannelIdentification(self, channelID):
        """!
        Check if channelID is not None

        @throws ChannelException if channelID is None.
        """
        if channelID is None:
            raise ChannelException("Phant channel must have identification")

class ChannelException:
    """!
    Channel related exceptions.
    """



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import threading
import time
import queue
import sys
from mqspeak.data import Measurement

class ChannnelUpdateSupervisor:
    """!
    Manage channel updaters. Object is responsible to delivering channel update event to
    correct Updater object.
    """

    ## @var channelUpdaterMapping
    # Mapping for {channel: updater}.

    def __init__(self, channelUpdaterMapping):
        """!
        Initiate ChannnelUpdateSupervisor object.

        @param channelUpdaterMapping Mapping for {channel: updater}.
        """
        self.channelUpdaterMapping = channelUpdaterMapping

    def dataAvailable(self, channel, measurement):
        """!
        Deliver new update to correct Updater object.

        @param channel
        @param measurement
        """
        updater = self.channelUpdaterMapping[channel]
        updater.dataAvailable(measurement)

    def setDispatcher(self, dispatcher):
        """!
        Assign a dispatcher to all updaters.

        @param dispatcher
        """
        for updater in self.channelUpdaterMapping.values():
            updater.setDispatcher(dispatcher)

    def stop(self):
        """!
        Stop execution of all updaters.
        """
        for updater in self.channelUpdaterMapping.values():
            updater.stop()

class BaseUpdater:
    """!
    Updater object base class!
    """

    ## @var channel
    # Updated channel.

    ## @var updateInterval
    # Channel update interval.

    ## @var isUpdateRunning
    # Boolean variable to keep track if some update is currently running.

    ## @var lastUpdated
    # Last update time.

    ## @var updateLock
    # Mutual exclusion to running updates.

    ## @var dispatcher
    # Update dispatcher object.

    def __init__(self, channel, updateInterval):
        """!
        Initiate BaseUpdater object.

        @param channel Updated channel.
        @param updateInterval Update interval.
        """
        self.channel = channel
        self.updateInterval = updateInterval
        self.isUpdateRunning = False
        self.lastUpdated = datetime.datetime.min
        self.updateLock = threading.Semaphore(1)

    def setDispatcher(self, dispatcher):
        """!
        Assign a dispatcher.

        @param dispatcher Dispatcher object.
        """
        self.dispatcher = dispatcher

    def dataAvailable(self, measurement):
        """!
        Update new data.

        @param measurement
        """
        self.updateLock.acquire()
        self.handleAvailableData(measurement)
        self.updateLock.release()

    def runUpdate(self, measurement):
        """!
        Call this method in sub-class from handleAvailableData method.

        @param measurement
        """
        self.isUpdateRunning = True
        self.dispatcher.updateAvailable(self.channel, measurement, self)

    def runUpdateLocked(self, measurement):
        """!
        Run update. This method avoids race conditions. Do not call this method
        from handleAvailableData() metod - causes dead lock.

        @param measurement
        """
        self.updateLock.acquire()
        self.runUpdate(measurement)
        self.updateLock.release()

    def notifyUpdateResult(self, result):
        """!
        Callback method with update result

        @param result UpdateResult object.
        """
        self.updateLock.acquire()
        self.isUpdateRunning = False
        if result.wasSuccessful():
            self.restartUpdateIntervalCounter()
        self.resolveUpdateResult(result)
        self.updateLock.release()

    def stop(self):
        """!
        Override this method if updater manage some other running thread.
        """

    def isUpdateIntervalExpired(self):
        """!
        Check if Update interval has expired.

        @return True if update interval has expired, False otherwise.
        """
        return (datetime.datetime.now() - self.lastUpdated) > self.updateInterval

    def restartUpdateIntervalCounter(self):
        """!
        Restart interval counter.
        """
        self.lastUpdated = datetime.datetime.now()

    def handleAvailableData(self, measurement):
        """!
        Handle new data in updater. Override this mehod in sub-class.

        @param measurement New data measurement.
        """
        raise NotImplementedError("Override this mehod in sub-class")

    def resolveUpdateResult(self, result):
        """!
        Resolve update result in updater. Override this mehod in sub-class.

        @param result UpdateResult object.
        """
        raise NotImplementedError("Override this mehod in sub-class")

class BlackoutUpdater(BaseUpdater):
    """!
    Ignore any incomming data during blackkout period. Send first data after this
    period expires.
    """

    def handleAvailableData(self, measurement):
        """!
        @copydoc BaseUpdater::handleAvailableData()
        """
        if self.isUpdateIntervalExpired() and not self.isUpdateRunning:
            self.runUpdate(measurement)

    def resolveUpdateResult(self, result):
        """!
        @copydoc BaseUpdater::resolveUpdateResult()
        """
        pass

class SynchronousUpdater(BaseUpdater):
    """!
    Base class for all updaters which tries to update channel in synchronous fashion.
    """

    ## @var isUpdateScheduled
    # Boolean variable to track if some update is already scheduled.

    ## @var bufferLock
    # Mutual exclusion for measurement buffer.

    ## @var scheduleLock
    # Mutual exclusion for isUpdateScheduled flag.

    ## @var executors
    # Set of running executors.

    def __init__(self, channel, updateInterval):
        """!
        Initiate SynchronousUpdater object.

        @param channel
        @param updateInterval
        """
        BaseUpdater.__init__(self, channel, updateInterval)
        self.resetBuffer()
        self.isUpdateScheduled = False
        self.bufferLock = threading.Semaphore(1)
        self.scheduleLock = threading.Semaphore(1)
        # TODO: Check race conditions with this set.
        self.executors = set()

    def handleAvailableData(self, measurement):
        """!
        @copydoc BaseUpdater::handleAvailableData()
        """
        self.scheduleLock.acquire()
        if not self.isUpdateScheduled:
            if self.isUpdateRunning:
                # New data is available but update is still running.
                # Store data in buffer and schedule new update after current update is done.
                self.updateBuffer(measurement)
            else:
                self.runUpdate(measurement)
        else:
            # Update job is already scheduled. Just update buffer.
            self.updateBuffer(measurement)
        self.scheduleLock.release()

    def resolveUpdateResult(self, result):
        """!
        @copydoc BaseUpdater::resolveUpdateResult()
        """
        # Schedule new update job. Just for case that new data arrive before time
        # interval expires.
        self.scheduleLock.acquire()
        self.scheduleUpdateJob()
        self.scheduleLock.release()

    def updateBuffer(self, measurement):
        """
        Update buffered measurement.
        """
        self.bufferLock.acquire()
        self.storeUpdateData(measurement)
        self.bufferLock.release()

    def scheduleUpdateJob(self):
        """
        Schedule new update job.
        """
        self.isUpdateScheduled = True
        executor = SchedulerExecutor(
            datetime.timedelta(seconds=int(self.updateInterval.total_seconds())),
            self.onSchedule)
        threading.Thread(target=executor).start()
        self.executors.add(executor)

    def onSchedule(self, executor):
        """
        Callback method called when scheduler expires.
        """
        self.scheduleLock.acquire()
        self.executors.discard(executor)
        self.isUpdateScheduled = False
        if self.isDataBuffered():
            data = self.pullMeasurement()
            self.runUpdateLocked(data)
        self.scheduleLock.release()

    def pullMeasurement(self):
        """!
        Atomically get buferred measurement and clear buffer.

        @return Measurement.
        """
        d = None
        self.bufferLock.acquire()
        d = self.getMeasurement()
        self.resetBuffer()
        self.bufferLock.release()
        return d

    def stop(self):
        """!
        Stop all executors.
        """
        for executor in self.executors:
            executor.stop()
        self.executors = set()

    def resetBuffer(self):
        """!
        Clear internal buffer with measurements. Override this mehod in sub-class.
        """
        raise NotImplementedError("Override this mehod in sub-class")

    def storeUpdateData(self, measurement):
        """!
        Save new measurement. Override this mehod in sub-class.

        @param measurement
        """
        raise NotImplementedError("Override this mehod in sub-class")

    def isDataBuffered(self):
        """!
        Check if there are some stored measurements. Override this mehod in sub-class.
        """
        raise NotImplementedError("Override this mehod in sub-class")

    def getMeasurement(self):
        """!
        Get stored measurement. Override this mehod in sub-class.
        """
        raise NotImplementedError("Override this mehod in sub-class")

class BufferedUpdater(SynchronousUpdater):
    """!
    Implement some timer to send update is time elapses. Don't wait for incoming data
    after time expires.
    """

    ## @var measurement
    # Stored measurement.

    def resetBuffer(self):
        """!
        @copydoc SynchronousUpdater::resetBuffer()
        """
        self.measurement = None

    def storeUpdateData(self, measurement):
        """!
        @copydoc SynchronousUpdater::storeUpdateData()
        """
        self.measurement = measurement

    def isDataBuffered(self):
        """
        @copydoc SynchronousUpdater::isDataBuffered()
        """
        return self.measurement is not None

    def getMeasurement(self):
        """!
        @copydoc SynchronousUpdater::getMeasurement()
        """
        return self.measurement

class AverageUpdater(SynchronousUpdater):
    """!
    Like BufferedUpdater but keep track all data which wasn't send and calculate
    average value while sending them.
    """

    ## @var intervalMeasurements
    # Stored measurements to compute average values.

    def storeUpdateData(self, measurement):
        """!
        @copydoc SynchronousUpdater::storeUpdateData()
        """
        if self.isAllMeasurementValuesValid(measurement):
            self.intervalMeasurements.append(measurement)
        else:
            print("Can't convert all measured values to numbers: {}".format(measurement), file=sys.stderr)

    def resetBuffer(self):
        """!
        @copydoc SynchronousUpdater::resetBuffer()
        """
        self.intervalMeasurements = []

    def isDataBuffered(self):
        """!
        @copydoc SynchronousUpdater::isDataBuffered()
        """
        return len(self.intervalMeasurements) > 0

    def getMeasurement(self):
        """!
        @copydoc SynchronousUpdater::getMeasurement()
        """
        return self.createAverageMeasurement()

    def createAverageMeasurement(self):
        """!
        Create measurement from collected data during update interval period.
        """
        averageData = {}
        for measurement in self.intervalMeasurements:
            for dataIdentifier, value in measurement.fields.items():
                if dataIdentifier not in averageData:
                    averageData[dataIdentifier] = 0
                averageData[dataIdentifier] += float(value)
        for dataIdentifier, value in averageData.items():
            averageData[dataIdentifier] = averageData[dataIdentifier] / float(len(self.intervalMeasurements))
        lastTime = self.intervalMeasurements[-1]
        return Measurement(averageData, lastTime)

    def isAllMeasurementValuesValid(self, measurement):
        """!
        Check if all measurement data can be converted to floating point numbers.

        @param measurement Measurement object.
        """
        for dataIdentifier, value in measurement.fields.items():
            try:
                float(value)
            except ValueError as ex:
                return False
        return True

class OnChangeUpdater(BaseUpdater):
    """!
    Send every value change.
    """

    ## @var changeBuffer
    # Store data which needs to be send.

    def __init__(self, channel):
        """!
        Initiate OnChangeUpdater object.

        @param channel
        """
        BaseUpdater.__init__(self, channel)
        self.changeBuffer = []
        raise NotImplementedError("Not implemented yet")

    def handleAvailableData(self, measurement):
        """!
        @copydoc BaseUpdater::handleAvailableData()
        """
        raise NotImplementedError("Not implemented yet")

    def resolveUpdateResult(self, result):
        """!
        @copydoc BaseUpdater::resolveUpdateResult()
        """
        raise NotImplementedError("Not implemented yet")

class SchedulerExecutor:
    """!
    Execute scheduler object in separate thread.
    """

    ## @var event
    # Event object.

    ## @var scheduleTime
    # Schedule time.

    ## @var action
    # Scheduled action.

    def __init__(self, scheduleTime, action):
        """!
        Initiate scheduler executor.

        @param scheduleTime Timedelta object.
        @param action Callable object executed after schedule time expires. Action takes one argument,
            which is reference to this executor.
        """
        self.event = threading.Event()
        self.scheduleTime = scheduleTime
        self.action = action

    def __call__(self):
        """!
        Run schedule execution.
        """
        scheduleExpires = not self.event.wait(self.scheduleTime.total_seconds())
        if scheduleExpires:
            self.action(self)

    def stop(self):
        """!
        Stop scheduler execution.
        """
        self.event.set()



#!/usr/bin/env python3
# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from mqspeak.collecting import DataCollector
from mqreceive.receiving import BrokerThreadManager
from mqspeak.sending import ChannelUpdateDispatcher
from mqspeak.system import System
from mqspeak.updating import ChannnelUpdateSupervisor

def main():
    System.initialize()

    # Channel update dispatcher object
    channelConvertMapping = System.getChannelConvertMapping()
    updateDispatcher = ChannelUpdateDispatcher(channelConvertMapping)

    channelUpdateSupervisor = ChannnelUpdateSupervisor(System.getChannelUpdateMapping())
    channelUpdateSupervisor.setDispatcher(updateDispatcher)
    dataCollector = DataCollector(System.getUpdateBuffers(), channelUpdateSupervisor)

    # MQTT cliens
    brokerManager = BrokerThreadManager(System.getBrokerListenDescriptors(), dataCollector)

    # run all MQTT client threads
    brokerManager.start()

    # run main thread
    try:
        updateDispatcher.run()
    except KeyboardInterrupt as ex:

        # program exit
        channelUpdateSupervisor.stop()
        brokerManager.stop()
        updateDispatcher.stop()

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt as ex:
        pass



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from mqspeak.collecting import UpdateBuffer
from mqspeak.config import ProgramConfig, ConfigException
from mqspeak.data import MeasurementParamConverter
from mqspeak import args

class System:
    """!
    System initialization object. This object encapsulates program configuration state
    defined by command line arguments and parsed configuration file.
    """

    @classmethod
    def initialize(cls):
        """!
        Initiate system configuration.
        """
        cls.cliArgs = args.parse_args()
        config = ProgramConfig(cls.cliArgs.config)
        # TODO: handle config exceptions
        try:
            cls.configCache = config.parse()
        except ConfigException as ex:
            print("Configuration error: {}".format(ex), file=sys.stderr)
            exit(1)

        cls.verbose = cls.cliArgs.verbose

    @classmethod
    def getChannelConvertMapping(cls):
        """!
        Get mapping for converting measurements for each channel.

        @return {channel: channelParamConverter}
        """
        channelConvertMapping = {}
        for channel, _, updateMapping in cls.configCache.channelUpdateDescribtors:
            channelConvertMapping[channel] = MeasurementParamConverter(updateMapping)
        return channelConvertMapping

    @classmethod
    def getBrokerListenDescriptors(cls):
        """!
        Get list of tuples (broker, ["subscribeTopic"]).

        @return (broker, ["subscribeTopic"])
        """
        return cls.configCache.listenDescriptors

    @classmethod
    def getUpdateBuffers(cls):
        """!
        Get list of UpdateBuffer instances.

        @return
        """
        updateBuffers = []
        for channel, _, updateMapping in cls.configCache.channelUpdateDescribtors:
            dataIdentifiers = list(updateMapping.keys())
            updateBuffer = UpdateBuffer(channel, dataIdentifiers)
            updateBuffers.append(updateBuffer)
        return updateBuffers

    @classmethod
    def getChannelUpdateMapping(cls):
        """!
        Get mapping {channel: updater}.

        @return {channel: updater}
        """
        channelUpdateMapping = {}
        for channel, updater, _ in cls.configCache.channelUpdateDescribtors:
            channelUpdateMapping[channel] = updater
        return channelUpdateMapping



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from mqreceive.data import DataIdentifier
from mqspeak.data import Measurement

class DataCollector:
    """!
    Object for collecting received data from MQTT brokers. This object is also resposible
    to provide received data to each update buffer. If some of update buffers becomes full,
    its content is given to ChannelUpdateSupervisor object.
    """

    ## @var updateBuffers
    # List of UpdateBuffer objects.

    ## @var channelUpdateSupervisor
    # UpdateSupervisor object.

    def __init__(self, updateBuffers, channelUpdateSupervisor):
        """!
        Initiate DataCollector object.

        @param updateBuffers List of UpdateBuffer objects.
        @param channelUpdateSupervisor UpdateSupervisor object.
        """
        self.updateBuffers = updateBuffers
        self.channelUpdateSupervisor = channelUpdateSupervisor

    def onMessage(self, broker, topic, data):
        self.onNewData(DataIdentifier(broker, topic), data);

    def onNewData(self, dataIdentifier, data):
        """!
        Notify data collector when new data is available.

        @param dataIdentifier Data identification.
        @param data Payload.
        """
        for updateBuffer in self.updateBuffers:
            self.tryBuffer(updateBuffer, dataIdentifier, data)

    def tryBuffer(self, updateBuffer, dataIdentifier, data):
        """!
        Try to update buffer.

        @param updateBuffer Buffer to update.
        @param dataIdentifier Data identification.
        @param data Payload.
        """
        try:
            updateBuffer.updateReceivedData(dataIdentifier, data)
            if updateBuffer.isComplete():
                d = updateBuffer.getData()
                updateBuffer.reset()
                measurement = Measurement.currentMeasurement(d)
                self.channelUpdateSupervisor.dataAvailable(updateBuffer.channel, measurement)
        except TopicException as ex:
            pass

class UpdateBuffer:
    """!
    Object for buffering reqired data set before delivering them to ThingSpeak.
    """

    ## @var channel
    # Updated channel.

    ## @var dataIdentifiers
    # Iterable of DataIdentifier objects.

    ## @var dataDict
    # The {DataIdentifier: value} mapping.

    def __init__(self, channel, dataIdentifiers):
        """!
        Initiate UpdateBuffer object.

        @param channel Channel identification object.
        @param dataIdentifiers Iterable of DataIdentifier objects.
        """
        self.channel = channel
        self.dataIdentifiers = dataIdentifiers
        self.reset()

    def isComplete(self):
        """!
        Check if all required data are buffered.

        @return True if all required data are buffered, False otherwise.
        """
        return not any(x is None for x in self.dataDict.values())

    def updateReceivedData(self, dataIdentifier, value):
        """!
        Update received data.

        @param dataIdentifier Data identification.
        @param value Data content.
        @throws TopicException If unnessesary topic is updated.
        """
        if dataIdentifier not in self.dataDict:
            raise TopicException("Illegal topic update: {}".format(dataIdentifier))
        else:
            self.dataDict[dataIdentifier] = value

    def getData(self):
        """!
        Get dictionary with buffered data.

        @return Buffered data.
        """
        if not self.isComplete():
            raise TopicException("Some topic data is missing")
        else:
            return self.dataDict

    def reset(self):
        """!
        Clear buffered data.
        """
        self.dataDict = dict()
        for dataIdentifier in self.dataIdentifiers:
            self.dataDict[dataIdentifier] = None

    def __str__(self):
        """!
        Convert UpdateBuffer object to string.

        @return String.
        """
        return "UpdateBuffer({}: {})".format(self.channel, self.dataIdentifiers)

    def __repr__(self):
        """!
        Convert UpdateBuffer object to representation string.

        @return Representation string.
        """
        return "<{}>".format(self.__str__())

class TopicException(Exception):
    """!
    Update buffer related errors.
    """



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import datetime
from mqreceive.broker import Broker
from mqspeak.channel import ThingSpeakChannel, PhantChannel
from mqreceive.data import DataIdentifier
from mqspeak.updating import BlackoutUpdater, BufferedUpdater, AverageUpdater, OnChangeUpdater

class ProgramConfig:
    """!
    Program configuration parser.
    """

    ## @var configFile
    # Configuration file name.

    ## @var parser
    # Parser object.

    def __init__(self, configFile):
        """!
        Initiate program configuration object.

        @param configFile Path to configuration file.
        """
        self.configFile = configFile
        self.parser = configparser.ConfigParser()

    def parse(self):
        """!
        Parse config file.

        @return Configuration object.
        """
        self.parser.read(self.configFile)
        self.checkForMandatorySections()
        configCache = ConfigCache()
        for broker, subscriptions in self.getBrokers():
            configCache.addBroker(broker, subscriptions)
        for channel, updaterFactory, updateMappingFactory in self.getChannels():
            updater = updaterFactory.build(channel)
            updateMapping = updateMappingFactory.build(configCache)
            configCache.addChannel(channel, updater, updateMapping)
        return configCache

    def checkForMandatorySections(self):
        """!
        Check if all necessary sections are mandatory.

        @throws ConfigException if some section is missing.
        """
        self.checkForSectionList(["Brokers", "Channels"])

    def getBrokers(self):
        """!
        Get list of enabled brokers.

        @return Iterable of brokers.
        """
        section = "Brokers"
        self.checkForEnabledOption(section)
        brokerSections = self.parser.get(section, "Enabled").split()
        self.checkForSectionList(brokerSections)
        for brokerSection in brokerSections:
            self.checkForBrokerMandatoryOptions(brokerSection)
            broker = self.createBroker(brokerSection)
            subscriptions = self.getBrokerSubscribtions(brokerSection)
            yield broker, subscriptions

    def checkForBrokerMandatoryOptions(self, brokerSection):
        """!
        Check for mandatory options of broker section.

        @param brokerSection Broker section name.
        @throws ConfigException If some option is missing.
        """
        optionList = ["Topic"]
        self.checkForOptionList(brokerSection, optionList)

    def createBroker(self, brokerSection):
        """!
        Create broker object from broker section.

        @param brokerSection Broker section name.
        @return Broker object
        """
        try:
            options = self.parser.options(brokerSection)
            host = self.parser.get(brokerSection, "Host", fallback = "127.0.0.1")
            port = self.parser.getint(brokerSection, "Port", fallback = 1883)
            broker = Broker(brokerSection, host, port)
            if "user" in options or "password" in options:
                (user, password) = self.getBrokerCredentials(brokerSection)
                broker.setCredentials(user, password)
            return broker
        except ValueError as ex:
            raise ConfigException("Invalid broker port number: {}".format(self.parser.get(brokerSection, "Port")))

    def getBrokerCredentials(self, brokerSection):
        """!
        Get username and password of broker.

        @param brokerSection Broker section name.
        @return Tuple of (username, password).
        @throws ConfigException If username or password is missing in configuration file.
        """
        user = None
        password = None
        try:
            user = self.parser.get(brokerSection, "User")
        except configparser.NoOptionError as ex:
            raise ConfigException("Section {}: User option is missing".format(brokerSection))
        try:
            password = self.parser.get(brokerSection, "Password")
        except configparser.NoOptionError as ex:
            raise ConfigException("Section {}: Password option is missing".format(brokerSection))
        return user, password

    def getBrokerSubscribtions(self, brokerSection):
        """!
        Get list of broker subscribe topics.

        @param brokerSection Broker section name.
        @return List of broker subscribe topics.
        @throws ConfigException If zero topics are specified.
        """
        subscriptions = self.parser.get(brokerSection, "Topic").split()
        if len(subscriptions) == 0:
            raise ConfigException("At least one topic subscribe has to be defined")
        return subscriptions

    def getChannels(self):
        """!
        Create list of enable channels.

        @return Iterable of tuples of (channel, updaterFactory, updateMappingFactory).
        @throws ConfigException
        """
        section = "Channels"
        self.checkForEnabledOption(section)
        channelSections = self.parser.get(section, "Enabled").split()
        self.checkForSectionList(channelSections)
        for channelSection in channelSections:
            self.checkForChannelMandatoryOptions(channelSection)
            channel = self.createChannel(channelSection)
            updaterFactory = self.getChannelUpdater(channelSection)
            updateMappingFactory = self.getDataFieldMapping(channelSection)
            yield channel, updaterFactory, updateMappingFactory

    def checkForChannelMandatoryOptions(self, channelSection):
        """!
        Check if channel section has all mandatoty options.

        @param channelSection Channel section name.
        @throws ConfigException If some options are missing.
        """
        optionList = ["Key", "Type", "UpdateRate", "UpdateType", "UpdateFields"]
        self.checkForOptionList(channelSection, optionList)

    def createChannel(self, channelSection):
        """!
        Create channel object.

        @param channelSection Channel section name.
        @return Channel object.
        @throws ConfigException If configuration specifies unknown channel type.
        """
        channelID = self.parser.get(channelSection, "Id", fallback = None)
        writeKey = self.parser.get(channelSection, "Key")
        channelType = self.parser.get(channelSection, "Type")
        if channelType == "thingspeak":
            return ThingSpeakChannel(channelSection, channelID, writeKey)
        elif channelType == "phant":
            return PhantChannel(channelSection, channelID, writeKey)
        else:
            raise ConfigException("Unsupported channel type: {}".format(channelType))

    def getChannelUpdater(self, channelSection):
        """!
        Create channel updaterFactory.

        @param channelSection Channel section name.
        @return ChannelUpdaterFactory object for that channel.
        @throws ConfigException If channel specifies invalid update interval.
        """
        try:
            updateRate = datetime.timedelta(seconds = self.parser.getint(channelSection, "UpdateRate"))
            updaterName = self.parser.get(channelSection, "UpdateType")
            updaterCls, updaterArgs = self.createUpdaterFactory(updaterName, updateRate)
            return ChannelUpdaterFactory(updaterCls, updaterArgs)
        except ValueError as ex:
            raise ConfigException("Invalid update rate interval: {}".format(self.parser.get(channelSection, "UpdateRate")))

    def createUpdaterFactory(self, updaterName, updateRate):
        """!
        Create updater factory based on updater name.

        @param updaterName Updater name.
        @param updateRate Update interval.
        @return ChannelUpdaterFactory object.
        @throws ConfigException If unknown updater name is specified in config file.
        """
        updaterCls = None
        if updaterName == "blackout":
            updaterCls = BlackoutUpdater
        elif updaterName == "buffered":
            updaterCls = BufferedUpdater
        elif updaterName == "average":
            updaterCls = AverageUpdater
        elif updaterName == "onchange":
            updaterCls = OnChangeUpdater
        else:
            raise ConfigException("Unknown UpdateType: {}".format(updaterName))
        updaterArgs = (updateRate,)
        return updaterCls, updaterArgs

    def getDataFieldMapping(self, channelSection):
        """!
        Create field mapping for channel.

        @param channelSection Channel section name.
        @return Data field mapping.
        """
        updaterSection = self.parser.get(channelSection, "UpdateFields")
        return self.createDataFieldMapping(updaterSection)

    def createDataFieldMapping(self, updateSection):
        """!
        Create data field mapping from update section.

        @param updateSection Update section name.
        @return Data field mapping.
        """
        updateMappingFactory = UpdateMappingFactory()
        for mappingOption in self.parser.options(updateSection):
            optionValue = self.parser.get(updateSection, mappingOption).split()
            if len(optionValue) < 2:
                    raise ConfigException("{}: {} - option must contain two space separated values".format(updateSection, mappingOption))
            brokerName, topic = optionValue
            updateMappingFactory.addMapping(brokerName, topic, mappingOption)
        return updateMappingFactory

    def checkForEnabledOption(self, section):
        """!
        Check for "Enabled" option in given section.

        @param section Section name.
        @throws ConfigException If "Enabled" option is missing in section.
        """
        self.checkForOption(section, "Enabled")

    def checkForSectionList(self, sectionList):
        """!
        Check for list of sections in configuration file.

        @param sectionList
        @throws ConfigException
        """
        for section in sectionList:
            self.checkForSection(section)

    def checkForSection(self, section):
        """!
        Check for section in configuration file.

        @param section
        @throws ConfigException
        """
        if not self.parser.has_section(section):
            raise ConfigException("{} section is missing".format(section))

    def checkForOptionList(self, section, optionList):
        """!
        Check for list of options in single section.

        @param section
        @param optionList
        @throws ConfigException
        """
        for option in optionList:
            self.checkForOption(section, option)

    def checkForOption(self, section, option):
        """!
        Check for option in configuration file.

        @param section Section name.
        @param option Option name.
        @throws ConfigException If given section doesn!t contain specified option.
        """
        if not self.parser.has_option(section, option):
            raise ConfigException("Section {}: {} option is missing".format(section, option))

class ConfigCache:
    """!
    Cache object for storing app configuration.
    """

    ## @var listenDescriptors
    # Listen descriptors.

    ## @var channelUpdateDescribtors
    # Update descriptors.

    def __init__(self):
        """!
        Initiate configuration cache object.
        """
        self.listenDescriptors = []
        self.channelUpdateDescribtors = []

    def addBroker(self, broker, subscriptions):
        """!
        Add broker to configuration object.

        @param broker
        @param subscriptions
        """
        listenDescriptor = (broker, subscriptions)
        self.listenDescriptors.append(listenDescriptor)

    def addChannel(self, channel, updater, updateMapping):
        """!
        Add channel to configuration object.

        @param channel
        @param updater
        @param updateMapping
        @throws ConfigException When update mapping contains unknown broker.
        """
        channelUpdateDescribtor = (channel, updater, updateMapping)
        self.channelUpdateDescribtors.append(channelUpdateDescribtor)

    def check(self):
        """!
        @todo implement this method

        1. warning for unused brokers
        2. warning for not feasible topics in udate mappings.
        """

    def getBrokerByName(self, brokerName):
        """!
        Get broker object identified by its't name.

        @param @brokerName
        @throws ConfigException If the name don't match to any stored broker object.
        """
        for broker, subscriptions in self.listenDescriptors:
            if brokerName == broker.name:
                return broker
        raise ConfigException("Unknown broker name: {}".format(brokerName))

class ChannelUpdaterFactory:
    """!
    Build channel.
    """

    ## @var updaterCls
    # Updater class.

    ## @var updaterArgs
    # Arguments to call updater class.

    def __init__(self, updaterCls, updaterArgs):
        """!
        Initiate ChannelUpdaterFactory object.

        @param updaterCls
        @param updaterArgs
        """
        self.updaterCls = updaterCls
        self.updaterArgs = updaterArgs

    def build(self, channel):
        """!
        Build ChannelUpdater object.

        @param channel
        @return ChannelUpdater object.
        """
        return self.updaterCls(channel, *self.updaterArgs)

class UpdateMappingFactory:
    """!
    Build UpdateMapping object.
    """

    ## @var mapping
    # Mapping.

    def __init__(self):
        """!
        Initiate UpdateMappingFactory object.
        """
        self.mapping = {}

    def addMapping(self, brokerName, topic, field):
        """!
        Add build mapping.

        @param brokerName
        @param topic
        @param field
        """
        self.checkNewBrokerName(brokerName)
        self.mapping[brokerName].append((topic, field))

    def checkNewBrokerName(self, brokerName):
        """!
        Check if broker name exist in mapping. Create new if not.

        @param brokerName
        """
        if brokerName not in self.mapping:
            self.mapping[brokerName] = []

    def getNeededBrokers(self):
        """
        Get list of needed brokers.

        @return List of broker names.
        """
        return list(self.mapping.keys())

    def build(self, brokerNameResolver):
        """!
        Build UpdateMapping object.

        @param brokerNameResolver Object for resolving broker names into Broker objects.
        @return UpdateMapping object.
        """
        mapping = {}
        for brokerName in self.mapping.keys():
            broker = brokerNameResolver.getBrokerByName(brokerName)
            for topic, field in self.mapping[brokerName]:
                dataIdentifier = DataIdentifier(broker, topic)
                mapping[dataIdentifier] = field
        return mapping

class ConfigException(Exception):
    """!
    Exception raised during parsing configuration file
    """



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import paho.mqtt.client as mqtt
import socket
import sys
import threading
from mqreceive.data import DataIdentifier
from mqreceive.receiving import BaseBrokerReceiver
from mqspeak.system import System

class BrokerReceiver(BaseBrokerReceiver):
    """!
    Broker receiving thread
    """

    ## @var clientID
    # Client identification.

    ## @var broker
    # Broker object.

    ## @var subsciption
    # Broker subscribtion.

    ## @var dataCollector
    # Data collector object.

    ## @var client
    # MQTT client object.


    def __init__(self, clientID, listenDescriptor, dataCollector):
        """!
        Initiate BrokerReceiver object.

        @param clientID
        @param listenDescriptor Set containing following fields: (broker, subscription).
            @li broker: broker descriptor object
            @li subscription:
        @param dataCollector Listener object to deliver received updates.
        """
        self.clientID = clientID
        broker, subsciption = listenDescriptor
        self.broker = broker
        self.subsciption = subsciption
        self.dataCollector = dataCollector
        self.client = mqtt.Client(client_id = str(self.clientID))
        self._registerCallbacks()
        if self.broker.isAuthenticationRequired():
            self.client.username_pw_set(self.broker.user, self.broker.password)

    def _registerCallbacks(self):
        self.client.on_connect = self.onConnect
        self.client.on_disconnect = self.onDisconnect
        self.client.on_message = self.onMessage
        self.client.on_subscribe = self.onSubscribe
        self.client.on_unsubscribe = self.onUnsubscribe
        self.client.on_log = self.onLog

    def __call__(self):
        """!
        Broker receiver thread code.
        """
        keepAliveInterval = 60
        self.client.connect(self.broker.host, self.broker.port, keepAliveInterval)
        self.client.loop_forever()

    def onConnect(self, client, userdata, flags, rc):
        """!
        The callback for when the client receives a CONNACK response from the server.

        @param client
        @param userdata
        @param flags
        @param rc
        """
        print("{}: [{}] {}.".format(self._createClientIdentificationString(), rc, self._getClientConnectionStatus(rc)))
        for sub in self.subsciption:
            (result, mid) = self.client.subscribe(sub)

    def onDisconnect(self, client, userdata, rc):
        """!
        Callback when client is disconnected.

        @param client
        @param userdata
        @param rc
        """
        print("Client dicsconnect: {}".format(rc))

    def onMessage(self, client, userdata, msg):
        """!
        The callback for when a PUBLISH message is received from the server.

        @param client
        @param userdata
        @param msg
        """
        dataID = DataIdentifier(self.broker, msg.topic)
        try:
            data = msg.payload.decode("utf-8")
            self.dataCollector.onNewData(dataID, data)
        except UnicodeError as ex:
            print("Can't decode received message payload: {}".format(msg.payload), file=sys.stderr)

    def onSubscribe(self, client, userdata, mid, granted_qos):
        """!
        Callback when client is subscribed to topic.

        @param client
        @param userdata
        @param mid
        @param granted_qos
        """

    def onUnsubscribe(self, client, userdata, mid):
        """!
        Callback when client is unsubscribed.

        @param client
        @param userdata
        @param mid
        """

    def onLog(self, client, userdata, level, buf):
        """!
        Logging messages.

        @param client
        @param userdata
        @param level
        @param buf
        """
        if System.verbose:
            print("{}: [{}] {}".format(self._createClientIdentificationString(), level, buf))

    def stop(self):
        """!
        Stop receiver thread. Call this method to nicely end __call__() method.
        """
        self.client.disconnect()

    def _getClientConnectionStatus(self, rc):
        if rc == 0:
            return "Connection successful"
        elif rc == 1:
            return "Connection refused - incorrect protocol version"
        elif rc == 2:
            return "Connection refused - invalid client identifier"
        elif rc == 3:
            return "Connection refused - server unavailable"
        elif rc == 4:
            return "Connection refused - bad username or password"
        elif rc == 5:
            return "Connection refused - not authorised"
        else:
            return "Unknown return code: {}".format(rc)

    def _createClientIdentificationString(self):
        return "{} ({} [{}:{}])".format(self.clientID, self.broker.name, self.broker.host, self.broker.port)

class BrokerReceiverID:
    """!
    Broker receiver thread ID.
    """

    ## @var hostname
    # Machine hostname

    ## @var pid
    # Current process PID

    ## @var receiverID
    # Receiver thread ID.

    def __init__(self, hostname, pid, receiverID):
        """!
        Create broker receiver thread lient ID.

        @param hostname Machine hostname.
        @param pid Current PID.
        @param receiverID Receiving thread ID.
        """
        self.hostname = hostname
        self.pid = pid
        self.receiverID = receiverID

    def __str__(self):
        """!
        Convert BrokerReceiverID to string.

        @return String.
        """
        return "mqspeak-{}-{}-{}".format(self.hostname, self.pid, self.receiverID)

    def __repr__(self):
        """!
        Convert BrokerReceiverID to representation string.

        @return Representation string.
        """
        return "<{}>".format(self.__str__())

class BrokerReceiverIDManager:
    """!
    Manage receiver IDs.
    """

    ## @var hostname
    # Machine hostname

    ## @var pid
    # Current process PID

    ## @var receiverCounter
    # Keep track how many broker IDs were created.

    def __init__(self):
        """!
        Initiate BrokerReceiverIDManager object.
        """
        self.hostname = socket.gethostname()
        self.pid = os.getpid()
        self.receiverCounter = 0

    def createReceiverID(self):
        """!
        Create new receiver ID.

        @return Receiver ID.
        """
        _receiverID = self.receiverCounter
        self.receiverCounter += 1
        return BrokerReceiverID(self.hostname, self.pid, _receiverID)

class ThreadManagerException(Exception):
    """!
    Indicate BrokerThreadManager error.
    """



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import datetime
import http.client
import sys
import threading
import urllib.parse
from mqspeak.system import System
from mqspeak.channel import ChannelType

class ChannelUpdateDispatcher:
    """!
    Dispatching new update threads.
    """

    ## @var channelSenders
    # Channel sendres mapping.

    ## @var dispatchLock
    # Mutual exclusion for update job dispatching.

    ## @var running
    # Keep track if dispatcher is running.

    ## @var updateQueue
    # Queue of pending updates.

    def __init__(self, channelConvertMapping):
        """!
        Initiate ChannelUpdateDispatcher object.

        @param channelConvertMapping Mapping {channel: channelParamConverter}.
        """
        self.channelSenders = self.createChannelSenders(channelConvertMapping)
        self.dispatchLock = threading.Semaphore(0)
        self.running = False
        self.updateQueue = collections.deque()

    def createChannelSenders(self, channelConvertMapping):
        """!
        Crate channel senders mapping.

        @param channelConvertMapping ChannelConvertMapping object.
        @return Senders mapping.
        """
        channelSenders = {}
        channelSenders[ChannelType.thingspeak] = ThingSpeakSender(channelConvertMapping)
        channelSenders[ChannelType.phant] = PhantSender(channelConvertMapping)
        return channelSenders

    def updateAvailable(self, channel, measurement, resultNotify):
        """!
        Notify main thread when new data is available.

        @param channel
        @param measurement
        @param resultNotify
        """
        self.updateQueue.append((channel, measurement, resultNotify))
        self.dispatchLock.release()

    def sendJobDone(self, result):
        """!
        Notify updater.

        @param result
        """
        (returnCode, updater) = result
        updater.notifyUpdateResult(returnCode)

    def run(self):
        """!
        Start update dispatcher main loop.
        """
        self.running = True
        while self.running:
            self.dispatchLock.acquire()

            # check for stop() method call
            if not self.running:
                return

            # start send thread
            (channel, measurement, resultNotify) = self.updateQueue.popleft()
            self.dispatch(channel, measurement, resultNotify)

    def stop(self):
        """!
        Stop dispatcher thread.
        """
        self.running = False
        self.dispatchLock.release()

    def dispatch(self, channel, measurement, updater):
        """!
        Dispatch new ThingSpeak update thread.

        @param channel Updated channel.
        @param measurement Update data.
        @param updater Notified object with update results.
        """
        sendThread = threading.Thread(
            target = SendRunner(
                self.channelSenders[channel.channelType],
                channel,
                measurement,
                updater,
                self))
        sendThread.start()

class BaseSender:
    """!
    Sender base class.
    """

    ## @var channelConvertMapping
    # Mapping {channel: channelParamConverter}.

    def __init__(self, channelConvertMapping):
        """!
        Initiate Sender base class.

        @param channelConvertMapping Mapping {channel: channelParamConverter}.
        """
        self.channelConvertMapping = channelConvertMapping

    def send(self, channel,  measurement):
        """!
        Send measurement.

        @param channel
        @param measurement
        """
        try:
            status, reason, responseBytes = self.fetch(channel, measurement)
            response = self.decodeResponseData(responseBytes)
            if System.verbose:
                print("Channel {} response: {} {}: {}".format(channel, status, reason, response))
            result = (status, reason, response)
            success = self.checkSendResult(result)
            return UpdateResult(success)
        except BaseException as ex:
            print("Send exception: {}".format(ex), file = sys.stderr)
            return UpdateResult(False)

    def decodeResponseData(self, responseBytes):
        """!
        Decode response data.

        @param responseBytes
        @return Decoded data or decode error message.
        """
        data = None
        try:
            data = responseBytes.decode("utf-8").strip()
        except UnicodeError as ex:
            print("Can't decode response data: {}".format(responseBytes), file=sys.stderr)
            data = "<Decode error>"
        return data

    def fetch(self, channel, measurement):
        """!
        Upload data to channel. Override this mehod in sub-class.

        @param channel Channel identification object.
        @param measurement Uploaded data.
        """
        raise NotImplementedError("Override this mehod in sub-class")

    def checkSendResult(self, result):
        """!
        Check if data upload was succcessful. Override this mehod in sub-class.

        @param result Tuple of (status, reason, response).
        """
        raise NotImplementedError("Override this mehod in sub-class")

class ThingSpeakSender(BaseSender):
    """!
    Class for sending data to ThingSpeak. This class send measurements to URL api.thingspeak.com
    using HTTPS method. It also parses send result and checks if transfer was successful.
    """

    def fetch(self, channel, measurement):
        """!
        @copydoc BaseSender::fetch()
        """
        body = self.channelConvertMapping[channel].convert(measurement)
        body.update({'api_key': channel.apiKey})
        bodyEncoded = urllib.parse.urlencode(body)
        conn = http.client.HTTPSConnection("api.thingspeak.com")
        conn.request("POST", "/update", bodyEncoded)
        response = conn.getresponse()
        status = response.status
        reason = response.reason
        responseBytes = response.read()
        conn.close()
        return status, reason, responseBytes

    def checkSendResult(self, result):
        """!
        @copydoc BaseSender::checkSendResult()
        """
        (status, reason, data) = result
        if status != 200:
            print("Response status error: {} {} - {}.".format(status, reason, data), file = sys.stderr)
            return False
        elif data == "0":
            print("Data send error: ThingSpeak responded with return code 0.", file = sys.stderr)
            return False
        return True

class PhantSender(BaseSender):
    """!
    Send data to Phant server.
    """

    def fetch(self, channel, measurement):
        """!
        @copydoc BaseSender::fetch()
        """
        body = self.channelConvertMapping[channel].convert(measurement)
        bodyEncoded = urllib.parse.urlencode(body)
        headers = {"Phant-Private-Key": channel.apiKey,
                    "Content-Type": "application/x-www-form-urlencoded"}
        conn = http.client.HTTPConnection("data.sparkfun.com")
        conn.request("POST", "/input/{}".format(channel.channelID), bodyEncoded, headers=headers)
        response = conn.getresponse()
        status = response.status
        reason = response.reason
        responseBytes = response.read()
        conn.close()
        return status, reason, responseBytes

    def checkSendResult(self, result):
        """!
        @copydoc BaseSender::checkSendResult()
        """
        (status, reason, data) = result
        if status != 200:
            print("Response status error: {} {} - {}.".format(status, reason, data), file = sys.stderr)
            return False
        return True

class SendRunner:
    """!
    Callable wrapper class for sending data to ThingSpeak in separate thread.
    """

    ## @var sender
    # Sender object. This object must implement send(channel,  measurement) method.

    ## @var channel
    # Channel description object.

    ## @var measurement
    # Measured data.

    ## @var updater
    # Updater object which will be called by dispatcher after send job is done.

    ## @var jobNotify
    # Listener object called after data send.

    def __init__(self, sender, channel, measurement, updater, jobNotify):
        """!
        Initiate SendRunner object.

        @param sender Sender object. This object must implement send(channel,  measurement) method.
        @param channel Channel description object.
        @param measurement Measured data.
        @param updater Updater object which will be called by dispatcher after send job is done.
        @param jobNotify Listener object called after data send.
        """
        self.sender = sender
        self.channel = channel
        self.measurement = measurement
        self.updater = updater
        self.jobNotify = jobNotify

    def __call__(self):
        """!
        Thread code.
        """
        try :
            result = (self.sender.send(self.channel, self.measurement), self.updater)
            self.jobNotify.sendJobDone(result)
        except Exception as ex:
            self.jobNotify.sendJobDone(ex)

class UpdateResult:
    """!
    Encapsulate update result.
    """

    ## @var success
    # Flag if update was successful.

    def __init__(self, success):
        """!
        Initiate update result.

        @param success Indicate if update was successful or not
        """
        self.success = success

    def wasSuccessful(self):
        """!
        Check if UpdateResul was successful.

        @return True if was sucessful, False otherwise.
        """
        return self.success



# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import mqspeak

class HelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def __init__(self, prog, indent_increment=2, max_help_position=64, width=180):
        argparse.ArgumentDefaultsHelpFormatter.__init__(self, prog, indent_increment, max_help_position, width)

def create_parser():
    parser = argparse.ArgumentParser(
        description="MQTT bridge v{}".format(mqspeak.__version__),
        epilog="Copyright (C) {} <{}> {}".format(
            mqspeak.__author__,
            mqspeak.__email__,
            mqspeak.__url__),
        formatter_class=HelpFormatter)
    parser.add_argument('-c', '--config',
                        help='path to configuration file',
                        default="/etc/mqspeak.conf")
    parser.add_argument('-v', '--verbose',
                        help='verbose',
                        action='store_true')
    parser.add_argument('--version',
                        action='version',
                        version='{}'.format(mqspeak.__version__))
    return parser

def parse_args():
    parser = create_parser()
    return parser.parse_args()



#!/usr/bin/env python3
#
# Copyright (C) Ivo Slanina <ivo.slanina@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

import mqspeak

setup(
    name = "mqspeak",
    url = 'https://github.com/buben19/mqspeak',
    version = mqspeak.__version__,
    packages = find_packages(exclude = ['doc']),
    install_requires = ['mqreceive>=0.1.0'],
    author = mqspeak.__author__,
    author_email = mqspeak.__email__,
    description = "MQTT bridge",
    long_description = "Tool for collecting MQTT traffic and sending it to ThingSpeak or Phant servers.",
    license = "GPLv3",
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Customer Service',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Communications',
        'Topic :: Home Automation',
        'Topic :: Internet',
    ],
    keywords = 'iot internetofthings mqopen mqtt sensors thingspeak phant',
    entry_points = {
        "console_scripts": [
            "mqspeak = mqspeak.__main__:main"
        ]
    }
)
